#!/usr/bin/python

'''

HaloStack v0.2.1 by Panu Lahtinen (August 2013)

For more information, see http://www.iki.fi/~pnuu/halostack/

'''

from PythonMagick import Image as PMImage
from PythonMagick import Blob 
import numpy as np
from argparse import ArgumentParser
import os.path
import sys
import time
from glob import glob
import math
from scipy import ndimage
import matplotlib.pyplot as plt
# Import alignment code
from align_functions import *

VERSION = "0.2.1"

def IMtoNumpy(img, usm=None, verbose=False):
    if verbose:
        print "Converting image to numpy array"

    # Check if USM sharpening should be used
    if usm != None:
        if verbose:
            print "Running unsharp mask filter"
        r,s,a,t = (usm)
        img.unsharpmask(r,s,a,t)

    # Convert to RGB and get the pixel data as characters
    if verbose:
        print "Converting to RGB"
    img.magick('RGB')
    blb = Blob()
    img.write(blb)
    data = blb.data
    # Check what bit-depth the image has and convert to Numpy array
    if img.depth() == 8:
        n_im = np.fromstring(data, dtype='uint8') #.astype('uint16')
    else:
        n_im = np.fromstring(data, dtype='uint16')

    # Reshape the image from vector to height x width x channels
    h = img.rows()
    w = img.columns()
    c = 3
    if img.monochrome():
        c = 1

    n_im = n_im.reshape(h, w, c)
    return n_im

def NumpytoIM(img, usm=None, verbose=False):
    if verbose:
        print "Converting numpy array to ImageMagick"

    out_img = PMImage()
    if img.dtype == 'uint16':
        out_img.depth(16)
    else:
        out_img.depth(8)
    out_img.magick('RGB')
    h, w, c = img.shape
    size_str = str(w)+'x'+str(h)
    out_img.size(size_str)

    b = Blob()
    b.data = img.tostring()
    out_img.read(b)
    out_img.magick('PNG')

    # Check if USM sharpening should be used
    if usm != None:
        if verbose:
            print "Running unsharp mask filter"
        r, s, a, t = (usm)
        out_img.unsharpmask(r, s, a, t)

    return out_img

def parse_usm(usm, u):
    usm = usm.split(',')
    if len(usm) < 2 or len(usm) > 4:
        print "Unsharp mask parameters should be given as:"
        print "-%s radius,amount (-%s 50,4) OR" % (u,u)
        print "-%s radius,amount,sigma (-%s 30,1.1,10) OR" % (u,u)
        print "-%s radius,amount,sigma,threshold (-%s 25,2.1,5,10)" % (u,u)
        print "If omitted, sigma defaults to radius/2 and threshold defaults to 0"
        sys.exit()
    # Radius
    r = int(usm[0])
    # Amount
    a = float(usm[1])
    # Sigma
    if(len(usm) > 2):
        s = int(usm[2])
    else:
        s = r/2
    # Threshold
    if(len(usm) > 3):
        t = int(usm[3])
    else:
        t = 0

    return (r,s,a,t)


def get_image_coordinates(img_in, n, gamma=None):

    img = img_in.copy()

    if img.dtype == 'uint16':
        if gamma is not None:
            img = img.astype('float64')/2**16
            img **= (1/gamma)
            img *= 2**16
        img /= 256
        img[img > 255] = 255
        img = img.astype('uint8')

    fig = plt.figure()
    plt.imshow(img)
    fig.tight_layout()
    points = plt.ginput(n, show_clicks=True, timeout=0)
    plt.close()

    xs, ys = [], []
    for pnt in points:
        x, y = pnt
        xs.append(int(x))
        ys.append(int(y))

    return xs, ys


if __name__ == '__main__':

    start_time = time.time()

    parser = ArgumentParser(description='Halostack', prog="HaloStack")
    parser.add_argument("-a", "--average-stack", dest="avg_stack_file", 
                        default=None, metavar="FILE",
                        help="Filename of the average stack [default: not saved]")
    parser.add_argument("-m", "--max-intensity-stack", dest="max_int_stack_file", 
                        default=None, metavar="FILE",
                        help="Filename of the maximum intensity stack [default: not saved]")
    parser.add_argument("-M", "--max-deviation-stack", dest="max_dev_stack_file", 
                        default=None, metavar="FILE",
                        help="Filename of the maximum RGB-deviation stack [default: not saved]")
    parser.add_argument("-c", "--max-color-diff", dest="max_col_diff_stack_file",
                        default=None, metavar="FILE",
                        help="Filename of the maximum color separation stack [default: not saved]")
    parser.add_argument("-e", "--median", dest="median_stack_file", default=None,
                        metavar="FILE", help="Filename of the median stack [default: not saved]")
    parser.add_argument("-u", "--pre-usm", dest="pre_usm", default=None,
                        metavar="R,A[,S[,T]]", help="Unsharp mask parameters applied before stacking [default: None]")
    parser.add_argument("-U", "--post-usm", dest="post_usm", default=None,
                        metavar="R,A[,S[,T]]", help="Unsharp mask parameters applied after stacking [default: None]")
    parser.add_argument("-b", "--base-image", dest="base_image", default=None,
                        metavar="FILE", help="Use this image file as base for alignment and image intensity normalization")
    parser.add_argument("-n", "--normalize", dest="normalization", type=int, default=None,
                        metavar="NUM", help="Normalize images to given value, or to first image (-n 0) before stacking [default: no normalization]")
    parser.add_argument("-N", "--normalize-area", dest="normalization_area",
                        default=None, metavar="X1,Y1,X2,Y2", help="Area used to determine the normalization [default: full image]")
    parser.add_argument("-r", "--align-reference", dest="align_ref_loc",
                        default=None, metavar="x,y,s", help="Center and size of square used as reference area for image alignment [default: no image alignment]")
    parser.add_argument("-s", "--align-search-area", dest="align_search_area",
                        default=None, metavar="X1,Y1,X2,Y2", help="Area to be searched for image alignment point")
    parser.add_argument("-t", "--correlation-threshold", dest="correlation_threshold", 
                        default=0.7, metavar="NUM", type=float, help="Minimum alignment correlation required to include the image for stacking [default: 0.7]")
    parser.add_argument("-R", "--rotate", dest="rotate",
                        default=False, action="store_true", help="Co-rotate images.")
    parser.add_argument("-O", "--out-postfix", dest="out_postfix",
                        default=None, metavar="STR", help="Save images as PNG with given filename postfix [default: not saved]")
    parser.add_argument('fname_in', metavar="FILE", type=str, nargs='+',
                        help='list of files to be stacked')
    parser.add_argument("-v", "--version", version='%(prog)s '+VERSION, action="version")
    parser.add_argument("-V", "--verbose", dest="verbose", default=False,
                        action="store_true", help="Enable verbose output")
#    parser.add_argument("-C", "--click", dest="click", default=False,
#                        metavar="ref,srch,rot,norm",
#                        help="List of pixel coordinates selected with mouse")
    parser.add_argument("-d", "--no-alignment", dest="no_alignment", 
                        default=False, action="store_true", 
                        help="Do not ask control points even if they are not given at command-line")
    parser.add_argument("-g", "--view-gamma", dest="view_gamma", default=None, type=float, metavar="GAMMA", help="Adjust alignment preview gamma")

    # Parse commandline inputs
    args = vars(parser.parse_args())

    if args['normalization'] is not None:
        if args['normalization_area'] is not None:
            args['normalization_area'] = args['normalization_area'].split(',')
            if len(args['normalization_area']) != 4:
                print "Wrong number of parameters for normalization area."
                print "Use -E x1,y1,x2,y2 eg. -E 494,234,504,247"
                sys.exit()
            x1, y1, x2, y2 = args['normalization_area']
            args['normalization_area'] = (int(x1), int(y1), int(x2), int(y2))
            
    if args['align_ref_loc'] is not None:
        args['align_ref_loc'] = args['align_ref_loc'].split(',')
        if len(args['align_ref_loc']) != 3:
            print "Wrong number of parameters for alignment reference location and size."
            print "Use -r x,y,d  eg. -r 430,234,7  which defines a square from 423,227 to 437,241"
            sys.exit()
        args['align_ref_loc'] = [int(args['align_ref_loc'][0]), 
                                 int(args['align_ref_loc'][1]),
                                 int(args['align_ref_loc'][2])]

    if args['align_search_area'] is not None:
        args['align_search_area'] = args['align_search_area'].split(',')
        if len(args['align_search_area']) != 4:
            print "Wrong number of parameters for alignment search area."
            print "Use -s x1,y1,x2,y2 eg. -s 494,234,504,247"
            sys.exit()
        sx1, sy1, sx2, sy2 = args['align_search_area']
        args['align_search_area'] = (int(sx1), int(sy1), int(sx2), int(sy2))
    if args['pre_usm'] is not None:
        args['pre_usm'] = parse_usm(args['pre_usm'], 'u')
    if args['post_usm'] is not None:
        args['post_usm'] = parse_usm(args['post_usm'], 'U')
    args['rotate_points'] = None
    if args['view_gamma'] is not None:
        args['view_gamma'] = float(args['view_gamma'])


#    if args['click'] is not None:
#        # ref,srch,rot,norm
#        args['click'] = args['click'].split(',')
#        xxxxxx
#        fig = plt.figure()
#        plt.imshow(img_numpy)
#        fig.tight_layout()
#        pts = plt.ginput(n)


    frames = []

    # Ensure that all files are used also on Windows, as the command
    # prompt does not automatically parse wildcards to a list of images
    for f in args['fname_in']:
        if '*' in f:
            f_all = glob(f)
            for f2 in f_all:
                frames.append(f2)
                if args['verbose']:
                    print "Added %s to the image list" % f2
        else:
            frames.append(f)
            if args['verbose']:
                print "Added %s to the image list" % f

    args['fname_in'] = frames

    # If nothing is to be saved, or no input images are supplied, abort
    if args['avg_stack_file'] is None \
            and args['max_int_stack_file'] is None \
            and args['max_dev_stack_file'] is None \
            and args['max_col_diff_stack_file'] is None \
            and args['median_stack_file'] is None:
        print "No stacks will be saved, aborting."
        print ""
        parser.print_help()
        sys.exit()
    n_files = len(args['fname_in'])
    if n_files == 0:
        print "No input images, aborting."
        print ""
        parser.print_help()
        sys.exit()

    print "%d files found" % n_files

    # Set the named file as first entry in the image list
    # This file will be the base for alignment and intensity normalization
    if args['base_image'] is not None:
        idx = frames.index(args['base_image'])
        f = frames.pop(idx)
        frames.insert(0, f)
    else:
        args['base_image'] = args['fname_in'][0]

    if args['verbose']:
        print "%s will be used as base image" % args['base_image']

    # Determine image size
    base_img_fname = args['base_image']
    if args['verbose']:
        print "Reading %s" % base_img_fname
    img = PMImage(base_img_fname)
    w, h = img.size().width(), img.size().height()
    if img.monochrome():
        c = 1
    else:
        c = 3

    print "Detected image size is %d x %d (width x height)" % (w, h)

    print "%s will be used as base image" % args['base_image']
    img = IMtoNumpy(img, args['pre_usm'], verbose=args['verbose'])

    # If no alignment data are given ask user what is wanted, if any,
    # and get the pixel coordinates from mouse input
    if args['no_alignment'] is False and args['align_ref_loc'] is None:

        print "Click tight area (two cornerpoints) for reference location."
        xs, ys = get_image_coordinates(img, 2, gamma=args['view_gamma'])
        args['align_ref_loc'] = (int(np.ceil(np.mean(xs))), 
                                 int(np.ceil(np.mean(ys))), 
                                 int(np.ceil(np.min([np.abs(xs[0]-xs[1]), 
                                                     np.abs(ys[0]-ys[1])]
                                                    )
                                             )
                                     )
                                 )
        if args['align_search_area'] is None:
            print "Click corner points for the area where alignment reference will be in every image.\n"
            xs, ys = get_image_coordinates(img, 2, gamma=args['view_gamma'])
            args['align_search_area'] = (np.min(xs), np.min(ys),
                                         np.max(xs), np.max(ys))

        if args['rotate'] is True and args['rotate_points'] is None:
            print "Click corner points of two areas that will be rotated to be horizontal.\n"
            xs, ys = get_image_coordinates(img, 4, gamma=args['view_gamma'])
            x1 = int(np.ceil(np.mean(xs[:2])))
            x2 = int(np.ceil(np.mean(xs[2:])))
            y1 = int(np.ceil(np.mean(ys[:2])))
            y2 = int(np.ceil(np.mean(ys[2:])))
            rs = np.max(np.abs([xs[0]-xs[1],
                                xs[2]-xs[3],
                                ys[0]-ys[1],
                                ys[2]-ys[3]]))
            args['rotate_points'] = (x1, y1, x2, y2, rs)

    # Get data used for aligning images
    align_ref_data = None
    if args['align_ref_loc'] is not None:
        xc = args['align_ref_loc'][0]
        yc = args['align_ref_loc'][1]
        d = args['align_ref_loc'][2]
        align_ref_data = img[yc-d:yc+d+1, xc-d:xc+d+1,:].mean(axis=2)

    # Get data used to co-rotate the images
    if args['rotate_points'] is not None:
        rx1, ry1, rx2, ry2, rs = args['rotate_points']
        rot1_data = img[ry1-rs:ry1+rs+1,rx1-rs:rx1+rs+1,:].mean(axis=2)
        rot2_data = img[ry2-rs:ry2+rs+1,rx2-rs:rx2+rs+1,:].mean(axis=2)
        rotate_angle = 180*math.atan2((ry2-ry1),(rx2-rx1))/math.pi
        if rotate_angle is not 0:
            print "Rotation angle is %3.3f deg" % rotate_angle
            img = ndimage.rotate(img, rotate_angle, reshape=False)
            # If the base-image got rotated, we need to align it so
            # that the reference point is in the correct place
            b, xb, yb = align(img.mean(axis=2), 
                              align_ref_data, 
                              area=args['align_search_area'])
            tx = args['align_ref_loc'][0] - xb # translation in x-direction
            ty = args['align_ref_loc'][1] - yb # translation in y-direction
            print "Shift: x: %d, y: %d, correlation (R^2): %1.3f" % (tx, ty, b)
            n_x1, n_x2, n_y1, n_y2, o_x1, o_x2, o_y1, o_y2 = \
                calc_align_area(tx, ty, w, h) 
            # Move image data by the calculated amount
            n_img = 0*img.copy()
            n_img[n_y1:n_y2, n_x1:n_x2, :] = img[o_y1:o_y2, o_x1:o_x2, :]
            img = n_img.copy()

    # Calculate normalization value, or normalize the base image to given value
    if args['normalization'] is not None:
        if args['normalization_area'] is None:
            args['normalization_area'] = (0, 0, w, h)
        x1, y1, x2, y2 = args['normalization_area']

        if args['normalization'] > 0:
            if args['verbose']:
                print "Normalizing to %d" % args['normalization']
            img_tmp = img*(1.0*args['normalization']/
                           np.median(img[y1:y2, x1:x2, :].mean(axis=2)))
            if img.dtype == 'uint16':
                img_tmp[img_tmp > 2**16 - 1] = 2**16 - 1
            else:
                img_tmp[img_tmp > 2**8 - 1] = 2**8 - 1
            img = img_tmp.astype(img.dtype)
            del img_tmp

        if args['normalization'] == 0:
            if args['verbose']:
                print "Calculating normalization value"
            args['normalization'] = np.median(img[y1:y2, x1:x2, :].mean(axis=2))

    # Save image as PNG
    if args['out_postfix'] is not None:
        out_fname = '.'.join(base_img_fname.split('.')[0:-1]) # get rid of extension
        out_fname = "%s_%s.png" % (out_fname, args['out_postfix'])
        print "Saving image: %s" % out_fname
        out_img = NumpytoIM(img, None)
        out_img.write(out_fname)

    # Create output arrays for selected image stacks
    if args['verbose']:
        print "Initialize stacks"
    avg_stack = None
    max_int_stack = None
    max_dev_stack = None
    max_col_diff_stack = None
    median_stack = None
    if args['avg_stack_file'] is not None:
        avg_stack = np.zeros((h,w,c), dtype='double')
        avg_stack = avg_stack + img
    if args['max_int_stack_file'] is not None:
        max_int_stack = np.zeros((h,w,c), dtype=img.dtype)
        max_int_stack = max_int_stack + img
        # Maximum intensities need to be 32-bit to prevent overflow
        max_pix_intensities = np.zeros((h,w), dtype='uint32')
        max_pix_intensities = max_pix_intensities + img.sum(axis=2)
    if args['max_dev_stack_file'] is not None:
        max_dev_stack = np.zeros((h,w,c), dtype=img.dtype)
        max_dev_stack = max_dev_stack + img
        max_pix_deviations = np.zeros((h,w), dtype='uint16')
        max_pix_deviations = max_pix_deviations + img.std(axis=2)
    if args['max_col_diff_stack_file'] is not None:
        max_col_diff_stack = np.zeros((h,w,c), dtype=img.dtype)
        max_col_diff_stack = max_col_diff_stack + img
        max_col_diffs = img.max(axis=2)-img.min(axis=2)
    if args['median_stack_file'] is not None:
        # Import HDF5 library
        import h5py
        median_stack = np.zeros((h,w,c), dtype=img.dtype)
        # Create a HDF5 file for temporary storage space for median stack
        # calculations. The file is memorymapped, so access from the
        # program is easy, but amount of disk-I/O will be huge.
        median_hdf_file = args['median_stack_file'][:-3]
        median_hdf_file = median_hdf_file + 'H5'
        median_hdf_fid = h5py.File(median_hdf_file,'w')
        # Datasets will be h-high, w-wide and n_files-deep with the same
        # bit-depth as the input images
        hdf_R = median_hdf_fid.create_dataset('R', 
                                              (h, w, n_files), 
                                              dtype=img.dtype)
        hdf_G = median_hdf_fid.create_dataset('G', 
                                              (h, w, n_files), 
                                              dtype=img.dtype)
        hdf_B = median_hdf_fid.create_dataset('B', 
                                              (h, w, n_files), 
                                              dtype=img.dtype)
        # Put the data from the first image to the HDF5 file
        hdf_R[:,:,0] = img[:,:,0]
        hdf_G[:,:,0] = img[:,:,1]
        hdf_B[:,:,0] = img[:,:,2]

    # Remove the first frame from the list, as it has been already used
    args['fname_in'] = args['fname_in'][1:]

    frames_used = 1 # number of images used in the stacks
    for f in args['fname_in']:
        if os.path.isfile(f):
            print f
            # Read image to numpy array
            img = IMtoNumpy(PMImage(f), args['pre_usm'])

            # Calculate rotation and rotate the image
            if args['rotate_points'] is not None:
                rx1, ry1, rx2, ry2, rs = args['rotate_points']
                b1, x1b, y1b = align(img.mean(axis=2), 
                                     rot1_data, 
                                     area=(rx1-100,ry1-100,rx1+100,ry1+100))
                b2, x2b, y2b = align(img.mean(axis=2), 
                                     rot2_data, 
                                     area=(rx2-100,ry2-100,rx2+100,ry2+100))
                rotate_angle = 180*math.atan2((y2b-y1b), (x2b-x1b))/math.pi
                if rotate_angle is not 0:
                    print "Rotation angle: %3.3f deg with correlations of %1.3f and %1.3f" % (rotate_angle,b1,b2)
                    img = ndimage.rotate(img, rotate_angle, reshape=False)

            # Find alignment point and get translation parameters, adjust
            if align_ref_data is not None:
                b, xb, yb = align(img.mean(axis=2), 
                                  align_ref_data, 
                                  area=args['align_search_area'])
                tx = args['align_ref_loc'][0] - xb # translation in x-direction
                ty = args['align_ref_loc'][1] - yb # translation in y-direction

                print "Shift: x: %d, y: %d, correlation (R^2): %1.3f" % \
                    (tx, ty, b)

                # If correlation between alignment points is too low, skip
                # this frame
                if b < args['correlation_threshold']:
                    print "Correlation below threshold (%1.3f < %1.3f), skipping file %s" % (b, args['correlation_threshold'], f)
                    continue

                n_x1, n_x2, n_y1, n_y2, o_x1, o_x2, o_y1, o_y2 = \
                    calc_align_area(tx, ty, w, h) 

                # Move image data by the calculated amount
                n_img = 0*img.copy()
                n_img[n_y1:n_y2, n_x1:n_x2, :] = img[o_y1:o_y2, o_x1:o_x2, :]
                img = n_img.copy()
                del n_img

            # Normalize image intensity
            if args['normalization'] is not None:
                # Normalize the value within given image region. This has
                # the benefit that only the important area has any say on
                # the normalization, and the unimportant areas are ignored
                if args['verbose']:
                    print "Normalizing image intensity to %d" % \
                        args['normalization']
                x1, y1, x2, y2 = args['normalization_area']
                img_tmp = img*(1.0*args['normalization']/
                               np.median(img[y1:y2, x1:x2, :].mean(axis=2)))
                if img.dtype == 'uint16':
                    img_tmp[img_tmp > 2**16 - 1] = 2**16 - 1
                else:
                    img_tmp[img_tmp > 2**8 - 1] = 2**8 - 1
                img = img_tmp.astype(img.dtype)
                del img_tmp

            # Save image as png
            if args['out_postfix'] is not None:
                out_fname = '.'.join(f.split('.')[0:-1]) # get rid of extension
                out_fname = "%s_%s.png" % (out_fname, args['out_postfix'])
                print "Saving image: %s" % out_fname
                out_img = NumpytoIM(img, None)
                out_img.write(out_fname)
                del out_img

            # Add data to cumulative sum array, which will be used to
            # calculate average image
            if avg_stack is not None:
                if args['verbose']:
                    print "Updating average stack"
                avg_stack = avg_stack + img
        
            # Update pixels with greatest intensities
            if max_int_stack is not None:
                if args['verbose']:
                    print "Updating maximum intensity stack"
                intensities = img.sum(axis=2)
                idxs = intensities > max_pix_intensities
                max_pix_intensities[idxs] = intensities[idxs]
                max_int_stack[idxs, 0] = img[idxs, 0]
                max_int_stack[idxs, 1] = img[idxs, 1]
                max_int_stack[idxs, 2] = img[idxs, 2]
                del intensities
                del idxs
            # Update pixels with greatest standard deviation in color values
            if max_dev_stack is not None:
                if args['verbose']:
                    print "Updating maximum color deviation stack"
                deviations = img.std(axis=2)
                idxs = deviations > max_pix_deviations
                max_pix_deviations[idxs] = deviations[idxs]
                max_dev_stack[idxs, 0] = img[idxs, 0]
                max_dev_stack[idxs, 1] = img[idxs, 1]
                max_dev_stack[idxs, 2] = img[idxs, 2]
                del deviations
                del idxs
            # Update pixels with greatest separation between brightest and
            # dimmest color channel
            if max_col_diff_stack is not None:
                if args['verbose']:
                    print "Updating maximum color deviation stack"
                diff = img.max(axis=2) - img.min(axis=2)
                idxs = diff > max_col_diffs
                max_col_diffs[idxs] = diff[idxs]
                max_col_diff_stack[idxs, 0] = img[idxs, 0]
                max_col_diff_stack[idxs, 1] = img[idxs, 1]
                max_col_diff_stack[idxs, 2] = img[idxs, 2]
                del diff
                del idxs
            # Add image data to a HDF5 file used as data-cube. This file
            # is used as a temporary memory-mapped storage when
            # calculating median of all the co-aligned image pixels
            if median_stack is not None:
                if args['verbose']:
                    print "Updating median stack"
                hdf_R[:, :, frames_used] = img[:, :, 0]
                hdf_G[:, :, frames_used] = img[:, :, 1]
                hdf_B[:, :, frames_used] = img[:, :, 2]

            # Increment the frame counter. This happens only if the
            # correlation has been good enough in alignment
            frames_used += 1
            
        else:
            print "File %s does not exist, skipping" % f

    del img

    # Scale and save average stack
    if avg_stack is not None:
        if args['verbose']:
            print "Finishing average stack:"
            print "- subtract minimum value"
        # Image is scaled between 0 and 2^16 -1 to cover the whole 16-bit/color
        # range of the output image file
        avg_stack = avg_stack - avg_stack.min()
        if args['verbose']:
            print "- scale to 0 ... 65535"
        avg_stack = ((2**16-1)*avg_stack/avg_stack.max()).astype('uint16')
        avg_stack = NumpytoIM(avg_stack, args['post_usm'])
        if args['verbose']:
            print "- saving average stack"
        avg_stack.write(args['avg_stack_file'])
        print "%s saved" % args['avg_stack_file']

    # Save maximum intensity stack
    if max_int_stack is not None:
        max_int_stack = NumpytoIM(max_int_stack, args['post_usm'])
        max_int_stack.write(args['max_int_stack_file'])
        print "%s saved" % args['max_int_stack_file']

    # Save maximum color deviation stack
    if max_dev_stack is not None:
        max_dev_stack = NumpytoIM(max_dev_stack, args['post_usm'])
        max_dev_stack.write(args['max_dev_stack_file'])
        print "%s saved" % args['max_dev_stack_file']

    # Save maximum color separation stack
    if max_col_diff_stack is not None:
        max_col_diff_stack = NumpytoIM(max_col_diff_stack, args['post_usm'])
        max_col_diff_stack.write(args['max_col_diff_stack_file'])
        print "%s saved" % args['max_col_diff_stack_file']

    # Save median stack
    if median_stack is not None:
        print "Calculating median:"
        # Calculate median for valid data.
        if args['verbose']:
            print "- red channel"
        median_stack[:,:,0] = np.median(hdf_R[:, :, 0:frames_used], axis=2)
        if args['verbose']:
            print "- green channel"
        median_stack[:,:,1] = np.median(hdf_G[:, :, 0:frames_used], axis=2)
        if args['verbose']:
            print "- blue channel"
        median_stack[:,:,2] = np.median(hdf_B[:, :, 0:frames_used], axis=2)
        if args['verbose']:
            print "- closing HDF5 file %s" % median_hdf_file
        median_hdf_fid.close()
        print "Median stack calculated"
        median_stack = NumpytoIM(median_stack, args['post_usm'])
        if args['verbose']:
            print "Writing image"
        median_stack.write(args['median_stack_file'])
        print "%s saved" % args['median_stack_file']

    t = time.time() - start_time
    hours = int(t/3600)
    minutes = int((t-hours*3600)/60)
    seconds = int(t-hours*3600-minutes*60)
    print "All stacks created and saved from %d/%d images in %d h %d min %3.2f sec" % (frames_used, n_files, hours, minutes, seconds)
    print ""
